package type2;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import auxiliary.ListOfListContains;
import co_reachable.CoReachability;
import reachable.Reachability;

public class Type2Parser {

	public Map<String, List<List<String>>> createCNF(Map<String, List<List<String>>> rules, String start) {
		Set<String> nullable = nullableSet(rules);
		Map<String, List<List<String>>> newRules0 = new HashMap<>(rules);
		if (nullable.contains(start)) {
			List<List<String>> newStartRule = new LinkedList<>();
			List<String> c0 = new LinkedList<>();
			c0.add(start);
			newStartRule.add(c0);
			newStartRule.add(new LinkedList<>());
			newRules0.put("S'", newStartRule);
		}
		Map<String, List<List<String>>> newRules1 = removeUselessSymbols(newRules0,
				newRules0.containsKey("S'") ? "S'" : start);
		List<List<String>> startRule = newRules1.remove("S'");
		Map<String, List<List<String>>> newRules2 = removeEpsilonRules(newRules1, nullable);
		if (startRule != null) {
			newRules2.put("S'", startRule);
		}
		Map<String, List<List<String>>> newRules3 = normalize(newRules2);
		Map<String, List<List<String>>> newRules4 = removeChainRules(newRules3);
		Map<String, List<List<String>>> newRules5 = removeUselessSymbols(newRules4,
				newRules4.containsKey("S'") ? "S'" : start);
		return shortenConclusios(newRules5);
	}

	private Set<String> nullableSet(Map<String, List<List<String>>> rules) {
		Set<String> nullable = new HashSet<>();
		int n = nullable.size();
		int m = Integer.MAX_VALUE;
		while (n != m) {
			m = n;
			for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
				if (!nullable.contains(rule.getKey())) {
					for (List<String> conclusio : rule.getValue()) {
						boolean notNullable = false;
						for (String c : conclusio) {
							if (!nullable.contains(c)) {
								notNullable = true;
								break;
							}
						}
						if (!notNullable) {
							nullable.add(rule.getKey());
							break;
						}
					}
				}
			}
			n = nullable.size();
		}
		return nullable;
	}

	private Map<String, List<List<String>>> removeUselessSymbols(Map<String, List<List<String>>> rules, String start) {
		Map<String, List<List<String>>> rules1 = removeNonCoReachableVariables(rules, start);
		if (rules1.isEmpty()) {
			return rules1;
		}
		return removeNonReachableSymbols(rules1, start);
	}

	private Map<String, List<List<String>>> removeNonCoReachableVariables(Map<String, List<List<String>>> rules,
			String start) {
		Set<String> coReachable = CoReachability.coReachableSet(rules);
		if (!coReachable.contains(start)) {
			return new HashMap<>();
		}
		Map<String, List<List<String>>> newRules = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			if (coReachable.contains(rule.getKey())) {
				List<List<String>> newConclusios = new LinkedList<>();
				for (List<String> conclusio : rule.getValue()) {
					boolean containsNonCoReachable = false;
					for (String c : conclusio) {
						if (rules.containsKey(c) && !coReachable.contains(c)) {
							containsNonCoReachable = true;
							break;
						}
					}
					if (!containsNonCoReachable) {
						newConclusios.add(conclusio);
					}
				}
				newRules.put(rule.getKey(), newConclusios);
			}
		}
		return newRules;
	}

	private Map<String, List<List<String>>> removeNonReachableSymbols(Map<String, List<List<String>>> rules,
			String start) {
		Set<String> reachable = Reachability.reachableSet(rules, start);
		Map<String, List<List<String>>> newRules = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			if (reachable.contains(rule.getKey())) {
				List<List<String>> newConclusios = new LinkedList<>();
				for (List<String> conclusio : rule.getValue()) {
					boolean containsNonReachable = false;
					for (String c : conclusio) {
						if (rules.containsKey(c) && !reachable.contains(c)) {
							containsNonReachable = true;
							break;
						}
					}
					if (!containsNonReachable) {
						newConclusios.add(conclusio);
					}
				}
				newRules.put(rule.getKey(), newConclusios);
			}
		}
		return newRules;
	}

	private Map<String, List<List<String>>> removeEpsilonRules(Map<String, List<List<String>>> rules,
			Set<String> nullable) {
		Map<String, List<List<String>>> newRules = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			List<List<String>> newConclusios = new LinkedList<>();
			for (List<String> oldConclusio : rule.getValue()) {
				List<List<String>> newCons = removeNullable(oldConclusio, nullable);
				for (List<String> newCon : newCons) {
					if ((!newCon.isEmpty() && !ListOfListContains.contains(newConclusios, newCon))) {
						newConclusios.add(newCon);
					}
				}

			}
			newRules.put(rule.getKey(), newConclusios);
		}
		return newRules;
	}

	private List<List<String>> removeNullable(List<String> r, Set<String> nullable) {
		List<String> lNull = new LinkedList<>();
		List<Integer> posNull = new LinkedList<>();
		Map<Integer, String> posNonNull = new HashMap<>();
		for (int i = 0; i < r.size(); i++) {
			if (nullable.contains(r.get(i))) {
				lNull.add(r.get(i));
				posNull.add(i);
			} else {
				posNonNull.put(i, r.get(i));
			}
		}
		List<List<Integer>> ps = powerset(lNull.size() - 1);
		List<Map<Integer, String>> newCons = new LinkedList<>();
		for (List<Integer> s : ps) {
			Map<Integer, String> newCon = new HashMap<>();
			for (int i : s) {
				newCon.put(posNull.get(i), lNull.get(i));
			}
			newCons.add(newCon);
		}
		List<List<String>> newConsAdded = new LinkedList<>();
		for (Map<Integer, String> newCon : newCons) {
			List<String> newConAdded = new LinkedList<>();
			for (int i = 0; i < r.size(); i++) {
				String s0 = newCon.get(i);
				String s1 = posNonNull.get(i);
				if (s0 != null) {
					newConAdded.add(s0);
				} else if (s1 != null) {
					newConAdded.add(s1);
				}
			}
			newConsAdded.add(newConAdded);
		}
		return newConsAdded;
	}

	private List<List<Integer>> powerset(int n) {
		List<List<Integer>> ps = new ArrayList<>();
		if (n < 0) {
			ps.add(new ArrayList<>());
		} else {
			for (List<Integer> s : powerset(n - 1)) {
				ps.add(s);
				List<Integer> s1 = new ArrayList<>(s);
				s1.add(n);
				ps.add(s1);
			}
		}
		return ps;
	}

	private Map<String, List<List<String>>> normalize(Map<String, List<List<String>>> rules) {
		Map<String, List<List<String>>> newRules = new HashMap<>();
		Map<String, String> vars = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			List<List<String>> newConclusios = new LinkedList<>();
			for (List<String> conclusio : rule.getValue()) {
				if (conclusio.size() > 1) {
					List<String> newConclusio = new LinkedList<>();
					for (int i = 0; i < conclusio.size(); i++) {
						String s = conclusio.get(i);
						if (!rules.containsKey(s)) {
							vars.computeIfAbsent(s, k -> "V" + k);
							newConclusio.add(vars.get(s));
						} else {
							newConclusio.add(s);
						}
					}
					newConclusios.add(newConclusio);
				} else {
					newConclusios.add(conclusio);
				}
			}
			newRules.put(rule.getKey(), newConclusios);
		}
		for (Entry<String, String> entry : vars.entrySet()) {
			List<List<String>> ll = new LinkedList<>();
			List<String> l = new LinkedList<>();
			l.add(entry.getKey());
			ll.add(l);
			newRules.put(entry.getValue(), ll);
		}
		return newRules;
	}

	private Map<String, List<List<String>>> removeChainRules(Map<String, List<List<String>>> rules) {
		Map<String, List<List<String>>> newRules = new HashMap<>();
		Map<String, Set<String>> chainProdMap = findChainProductions(rules);
		Map<String, List<List<String>>> chainRulesDeleted = deleteChainRules(rules);
		for (String premise : rules.keySet()) {
			List<List<String>> newConclusios = new LinkedList<>();
			for (String s : chainProdMap.get(premise)) {
				for (List<String> newConclusio : chainRulesDeleted.get(s)) {
					if (!ListOfListContains.contains(newConclusios, newConclusio)) {
						newConclusios.add(newConclusio);
					}
				}

			}
			newRules.put(premise, newConclusios);
		}
		return newRules;
	}

	private Map<String, List<List<String>>> deleteChainRules(Map<String, List<List<String>>> rules) {
		Map<String, List<List<String>>> newRules = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			List<List<String>> newConclusios = new LinkedList<>();
			for (List<String> conclusio : rule.getValue()) {
				if (conclusio.size() == 1 && rules.containsKey(conclusio.get(0))) {
					continue;
				}
				newConclusios.add(conclusio);
			}
			newRules.put(rule.getKey(), newConclusios);
		}
		return newRules;
	}

	private Map<String, Set<String>> findChainProductions(Map<String, List<List<String>>> rules) {
		Map<String, Set<String>> chainProdMap = new HashMap<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			Set<String> chainProds = new HashSet<>();
			chainProds.add(rule.getKey());
			findChainProductions(rules, rule.getKey(), chainProds);
			chainProdMap.put(rule.getKey(), chainProds);
		}
		return chainProdMap;
	}

	private void findChainProductions(Map<String, List<List<String>>> rules, String key, Set<String> chainProds) {
		for (List<String> conclusio : rules.get(key)) {
			if (conclusio.size() == 1 && rules.containsKey(conclusio.get(0))
					&& !chainProds.contains(conclusio.get(0))) {
				chainProds.add(conclusio.get(0));
				findChainProductions(rules, conclusio.get(0), chainProds);
			}
		}
	}

	private Map<String, List<List<String>>> shortenConclusios(Map<String, List<List<String>>> rules) {
		Map<String, List<List<String>>> newRules = new HashMap<>();
		Map<String, List<List<String>>> addNewRules = new HashMap<>();
		int i = 0;
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			List<List<String>> newConclusios = new LinkedList<>();
			for (List<String> conclusio : rule.getValue()) {
				if (conclusio.size() > 2) {
					List<String> c0 = new LinkedList<>();
					c0.add(conclusio.get(0));
					c0.add("@" + i);
					newConclusios.add(c0);
					int k = conclusio.size() - 2;
					for (int j = 1; j < k; j++) {
						List<List<String>> csj = new LinkedList<>();
						List<String> cj = new LinkedList<>();
						cj.add(conclusio.get(j));
						cj.add("@" + (i + 1));
						csj.add(cj);
						addNewRules.put("@" + i++, csj);
					}
					List<List<String>> csn = new LinkedList<>();
					List<String> cn = new LinkedList<>();
					cn.add(conclusio.get(k));
					cn.add(conclusio.get(k + 1));
					csn.add(cn);
					addNewRules.put("@" + i++, csn);
				} else {
					newConclusios.add(conclusio);
				}
			}
			newRules.put(rule.getKey(), newConclusios);
		}
		newRules.putAll(addNewRules);
		return newRules;
	}

	private boolean cyk(Map<String, List<List<String>>> rules, String word, String start) {
		int n = word.length();
		String[][][] table = new String[n][n][rules.keySet().size()];
		Map<String, Integer> indexMap = indexMap(rules.keySet());
		for (int i = 0; i < n; i++) {
			Set<String> producesTerminal = producesTerminal(rules, word.charAt(i) + "");
			for (String premise : producesTerminal) {
				table[i][i][indexMap.get(premise)] = premise;
			}
		}
		for (int j = 1; j < n; j++) {
			for (int i = 0; i < n - j; i++) {
				Set<String> s = new HashSet<>();
				for (int k = i; k < i + j; k++) {
					Set<String> s0 = producesVariable(table[i][k], table[k + 1][i + j], rules);
					s.addAll(s0);
				}
				for (String v : s) {
					table[i][i + j][indexMap.get(v)] = v;
				}
			}
		}
		return table[0][n - 1][indexMap.get(start)] != null;
	}

	private Set<String> producesTerminal(Map<String, List<List<String>>> rules, String charAt) {
		Set<String> pt = new HashSet<>();
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			for (List<String> conclusio : rule.getValue()) {
				if (conclusio.size() == 1 && conclusio.get(0).equals(charAt)) {
					pt.add(rule.getKey());
				}
			}
		}
		return pt;
	}

	private Set<String> producesVariable(String[] v1, String[] v2, Map<String, List<List<String>>> rules) {
		Set<String> pv = new HashSet<>();
		for (String s1 : v1) {
			if (s1 != null) {
				for (String s2 : v2) {
					if (s2 != null) {
						pv.addAll(producesVariable(rules, s1, s2));
					}
				}
			}

		}
		return pv;
	}

	private Set<String> producesVariable(Map<String, List<List<String>>> rules, String v1, String v2) {
		Set<String> pv = new HashSet<>();
		List<String> varPair = new LinkedList<>();
		varPair.add(v1);
		varPair.add(v2);
		for (Entry<String, List<List<String>>> rule : rules.entrySet()) {
			for (List<String> conclusio : rule.getValue()) {
				if (conclusio.equals(varPair)) {
					pv.add(rule.getKey());
				}
			}
		}
		return pv;
	}

	private Map<String, Integer> indexMap(Set<String> vars) {
		Map<String, Integer> indexMap = new HashMap<>();
		int i = 0;
		for (String v : vars) {
			indexMap.put(v, i);
			i++;
		}
		return indexMap;
	}

	public boolean isPartOfLanguage(Map<String, List<List<String>>> cnfRules, String word, String start) {
		if (word.isEmpty()) {
			return ListOfListContains.contains(cnfRules.get(cnfRules.keySet().contains("S'") ? "S'" : start),
					new LinkedList<>());
		}
		return cyk(cnfRules, word, cnfRules.containsKey("S'") ? "S'" : start);
	}

	public static void main(String[] args) {
		Type2Parser parser2 = new Type2Parser();
		Map<String, List<List<String>>> rules = new HashMap<>();
		List<List<String>> s = List.of(List.of("C", "C", "a"), List.of("B"), List.of("c"));
		rules.put("S", s);
		List<List<String>> a = List.of(List.of("a", "C", "S"), List.of("b"), List.of("b", "D"));
		rules.put("A", a);
		List<List<String>> b = List.of(List.of(), List.of("S", "A"));
		rules.put("B", b);
		List<List<String>> c = List.of(List.of("B", "B", "A"), List.of("c", "B", "c"));
		rules.put("C", c);
		List<List<String>> d = List.of(List.of("c", "c", "D"), List.of("a", "c"), List.of("c"));
		rules.put("D", d);
		Map<String, List<List<String>>> cnfRules = parser2.createCNF(rules, "S");
		System.out.println(cnfRules);
		System.out.println(parser2.isPartOfLanguage(cnfRules,
				"bcacaaccccaaabacbccacbacbabbacbaabcabcbaaaacabacbaaabccccacacbccaccbcccaabacbabbaca", "S"));
	}

}
